from __future__ import annotations
from collections.abc import Generator
from pathlib import Path
from typing import Union
import cstag
import mappy


def revcomp(sequence: str) -> str:
    complement = {"A": "T", "C": "G", "G": "C", "T": "A"}
    return "".join(complement[nt] for nt in sequence[::-1])


def to_sam(path_reference_fasta: str, path_query_fastq: str, cslong: bool = True) -> Generator[str]:
    """Align seqences using mappy and Convert PAF to SAM

    Args:
        path_reference_fasta (str): Path of reference fasta
        path_query_fastq (str): Path of query fasta/fastq
        cslong (bool, optional): long formatted CS tag if True. Defaults to True

    Returns:
        list: List of SAM
    """
    # SQ header
    SAM = [f"@SQ\tSN:{n}\tLN:{len(s)}" for n, s, _ in mappy.fastx_read(path_reference_fasta)]
    # Mappy
    ref = mappy.Aligner(path_reference_fasta)
    if not ref:
        raise AttributeError(f"Failed to load f{path_reference_fasta}")
    for query_name, query_sequence, query_quality in mappy.fastx_read(path_query_fastq):
        for hit in ref.map(query_sequence, cs=True):
            # flag
            if hit.is_primary:
                flag = 0 if hit.strand == 1 else 16
            else:
                flag = 2048 if hit.strand == 1 else 2064
            # Append softclips to CIGAR
            cigar = hit.cigar_str
            if hit.q_st > 0:
                softclip = str(hit.q_st) + "S"
                cigar = softclip + cigar if hit.strand == 1 else cigar + softclip
            if len(query_sequence) - hit.q_en > 0:
                softclip = str(len(query_sequence) - hit.q_en) + "S"
                cigar = cigar + softclip if hit.strand == 1 else softclip + cigar
            # Revcomp
            if not hit.strand == 1:
                query_sequence = revcomp(query_sequence)
            query_sequence = query_sequence.upper()
            # cslong
            cs = "cs:Z:" + hit.cs
            if cslong:
                cs = cstag.lengthen(hit.cs, cigar, query_sequence)
            # summarize
            alignment = [
                query_name,
                flag,
                hit.ctg,
                hit.r_st + 1,
                hit.mapq,
                cigar,
                "*",
                0,
                0,
                query_sequence,
                query_quality,
                cs,
            ]
            alignment = [str(a) for a in alignment]
            SAM.append("\t".join(alignment))
    for record in SAM:
        yield record


def output_sam(TEMPDIR, path_fasta, name_fasta, path_fastq, name_fastq):
    sam = to_sam(str(path_fasta), path_fastq)
    output_sam = Path(TEMPDIR, "sam", f"{name_fastq}_{name_fasta}.sam")
    output_sam.write_text("\n".join(sam))


########################################################################
# Create faidx
########################################################################


def make_faidx(path_fasta: Union[Path, str]) -> str:
    fasta = Path(path_fasta).read_text().split()
    name = fasta[0].strip(">")
    length = len("".join(fasta[1:]))
    offset = len(fasta[0]) + 1
    linebase = len(fasta[1])
    linewidth = len(fasta[1]) + 1
    faidx = list(map(str, [name, length, offset, linebase, linewidth]))
    faidx = "\t".join(faidx) + "\n"
    return faidx



from __future__ import annotations
from .core import core
from .postprocess import report


def execute(arguments: dict[str]):
    core.execute(arguments)
    name = arguments["name"]
    report.report(name)



from __future__ import annotations
import hashlib
import shutil
import midsv
from pathlib import Path
from . import preprocess
from . import classification
from . import clustering
from . import consensus
from . import report


def execute(arguments: dict) -> None:
    SAMPLE = arguments["sample"]
    CONTROL = arguments["control"]
    ALLELE = arguments["allele"]
    NAME = arguments["name"]
    THREADS = arguments["threads"]
    try:
        GENOME = arguments["genome"]
    except KeyError:
        GENOME = False
        GENOME_COODINATES = None
        CHROME_SIZE = None

    ##########################################################
    # Check inputs
    ##########################################################
    preprocess.check_inputs.check_files(SAMPLE, CONTROL, ALLELE)
    TEMPDIR = Path("DAJINResults", ".tempdir", NAME)
    IS_CACHE_CONTROL = preprocess.check_inputs.is_cache_control(CONTROL, TEMPDIR)
    IS_CACHE_GENOME = preprocess.check_inputs.is_cache_genome(GENOME, TEMPDIR, IS_CACHE_CONTROL)
    UCSC_URL, GOLDENPATH_URL = None, None
    if GENOME and not IS_CACHE_GENOME:
        UCSC_URL, GOLDENPATH_URL = preprocess.check_inputs.check_and_fetch_genome(GENOME)

    ##########################################################
    # Format inputs
    ##########################################################
    SAMPLE_NAME = preprocess.format_inputs.extract_basename(SAMPLE)
    CONTROL_NAME = preprocess.format_inputs.extract_basename(CONTROL)
    DICT_ALLELE = preprocess.format_inputs.dictionize_allele(ALLELE)

    preprocess.format_inputs.make_directories(TEMPDIR)

    if GENOME:
        GENOME_COODINATES, CHROME_SIZE = preprocess.format_inputs.get_coodinates_and_chromsize(
            TEMPDIR, GENOME, DICT_ALLELE, UCSC_URL, GOLDENPATH_URL, IS_CACHE_GENOME
        )
        if not IS_CACHE_GENOME:
            preprocess.format_inputs.cache_coodinates_and_chromsize(TEMPDIR, GENOME, GENOME_COODINATES, CHROME_SIZE)

    ################################################################################
    # Export fasta files as single-FASTA format
    ################################################################################
    # TODO: use yeild, not export
    for identifier, sequence in DICT_ALLELE.items():
        contents = "\n".join([">" + identifier, sequence]) + "\n"
        output_fasta = Path(TEMPDIR, "fasta", f"{identifier}.fasta")
        output_fasta.write_text(contents)

    ###############################################################################
    # Mapping with minimap2/mappy
    ###############################################################################
    for path_fasta in Path(TEMPDIR, "fasta").glob("*.fasta"):
        name_fasta = path_fasta.stem
        if name_fasta not in set(DICT_ALLELE.keys()):
            continue
        if not IS_CACHE_CONTROL:
            preprocess.mappy_align.output_sam(TEMPDIR, path_fasta, name_fasta, CONTROL, CONTROL_NAME)
        preprocess.mappy_align.output_sam(TEMPDIR, path_fasta, name_fasta, SAMPLE, SAMPLE_NAME)

    if not GENOME:
        path_control = Path(TEMPDIR, "fasta", "control.fasta")
        faidx_control = preprocess.mappy_align.make_faidx(path_control)
        shutil.copy(path_control, Path(TEMPDIR, "report", ".igvjs"))
        Path(TEMPDIR, "report", ".igvjs", "control.fasta.fai").write_text(faidx_control)

    ########################################################################
    # MIDSV conversion
    ########################################################################
    if not IS_CACHE_CONTROL:
        for path_sam in Path(TEMPDIR, "sam").glob(f"{CONTROL_NAME}*"):
            preprocess.calc_midsv.output_midsv(TEMPDIR, path_sam, DICT_ALLELE)

    for path_sam in Path(TEMPDIR, "sam").glob(f"{SAMPLE_NAME}*"):
        preprocess.calc_midsv.output_midsv(TEMPDIR, path_sam, DICT_ALLELE)

    ###############################################################################
    # Cashe inputs (control)
    ###############################################################################

    if not IS_CACHE_CONTROL:
        control_hash = Path(CONTROL).read_bytes()
        control_hash = hashlib.sha256(control_hash).hexdigest()
        PATH_CACHE_HASH = Path(TEMPDIR, "cache", "control_hash.txt")
        PATH_CACHE_HASH.write_text(str(control_hash))

    ########################################################################
    # Classify alleles
    ########################################################################

    path_midsv = Path(TEMPDIR, "midsv").glob(f"{SAMPLE_NAME}*")
    classif_sample = classification.classify_alleles(path_midsv, SAMPLE_NAME)

    ########################################################################
    # Detect Structural variants
    ########################################################################

    for classif in classif_sample:
        classif["SV"] = classification.detect_sv(classif["CSSPLIT"], threshold=50)

    ########################################################################
    # Clustering
    ########################################################################

    MASKS_CONTROL = clustering.mask_control(TEMPDIR, DICT_ALLELE, CONTROL_NAME)

    diffloci_by_alleles = clustering.extract_different_loci(
        TEMPDIR, classif_sample, MASKS_CONTROL, DICT_ALLELE, CONTROL_NAME
    )
    clust_sample = clustering.add_labels(classif_sample, diffloci_by_alleles)
    clust_sample = clustering.add_readnum(clust_sample)
    clust_sample = clustering.add_percent(clust_sample)
    clust_sample = clustering.update_labels(clust_sample)

    ########################################################################
    # Consensus call
    ########################################################################

    RESULT_SAMPLE, cons_percentage, cons_sequence = consensus.call(clust_sample, diffloci_by_alleles, DICT_ALLELE)

    ########################################################################
    # Output Report：RESULT/FASTA/HTML/BAM/VCF
    ########################################################################
    # RESULT
    midsv.write_jsonl(RESULT_SAMPLE, Path(TEMPDIR, "result", f"{SAMPLE_NAME}.jsonl"))

    # FASTA
    for header, cons_seq in cons_sequence.items():
        cons_fasta = report.report_files.to_fasta(header, cons_seq)
        Path(TEMPDIR, "report", "FASTA", f"{SAMPLE_NAME}_{header}.fasta").write_text(cons_fasta)

    # HTML
    for header, cons_per in cons_percentage.items():
        cons_html = report.report_files.to_html(header, cons_per)
        Path(TEMPDIR, "report", "HTML", f"{SAMPLE_NAME}_{header}.html").write_text(cons_html)

    # BAM
    report.report_bam.output_bam(TEMPDIR, RESULT_SAMPLE, SAMPLE_NAME, GENOME, GENOME_COODINATES, CHROME_SIZE, THREADS)

    # VCF
    # working in progress




from __future__ import annotations
import midsv
from pathlib import Path
from copy import deepcopy
from collections import defaultdict
from itertools import groupby
from .make_scores import make_scores
from .return_labels import return_labels
from .screen_diffloci import screen_different_loci


def extract_different_loci(TEMPDIR, classif_sample, MASKS_CONTROL, DICT_ALLELE, CONTROL_NAME):
    """
    Extract significantly different base loci between Sample and Control
    """
    dict_cssplit_control = defaultdict(list[dict])
    for allele in DICT_ALLELE.keys():
        path_control = Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_{allele}.jsonl")
        cssplit_control = [cs["CSSPLIT"] for cs in midsv.read_jsonl(path_control)]
        dict_cssplit_control[allele] = cssplit_control
    classif_sample.sort(key=lambda x: (x["ALLELE"], x["SV"]))
    diffloci_by_alleles = defaultdict(list[dict])
    for (ALLELE, SV), group in groupby(classif_sample, key=lambda x: (x["ALLELE"], x["SV"])):
        cssplit_sample = [record["CSSPLIT"] for record in group]
        cssplit_control = dict_cssplit_control[ALLELE]
        sequence = DICT_ALLELE[ALLELE]
        masks_control = MASKS_CONTROL[ALLELE]
        diffloci = screen_different_loci(
            cssplit_sample, cssplit_control, sequence, masks_control, alpha=0.01, threshold=0.05
        )
        diffloci_by_alleles[f"{ALLELE}-{SV}"] = diffloci
    return diffloci_by_alleles


def add_labels(classif_sample, diffloci_by_alleles):
    labels = []
    label_start = 1
    classif_sample.sort(key=lambda x: (x["ALLELE"], x["SV"]))
    for (ALLELE, SV), group in groupby(classif_sample, key=lambda x: (x["ALLELE"], x["SV"])):
        key = f"{ALLELE}-{SV}"
        cssplit_sample = [g["CSSPLIT"] for g in group]
        diffloci = diffloci_by_alleles[key]
        scores = []
        if diffloci is not None:
            scores = make_scores(cssplit_sample, diffloci)
        if any(scores):
            labels += [label + label_start for label in return_labels(scores).tolist()]
        else:
            labels += [label_start] * len(cssplit_sample)
        label_start = len(set(labels)) + 1
    clust_sample = deepcopy(classif_sample)
    for clust, label in zip(clust_sample, labels):
        clust["LABEL"] = label
    return clust_sample


def add_readnum(clust_sample):
    readnum = defaultdict(int)
    for cs in clust_sample:
        readnum[cs["LABEL"]] += 1
    for cs in clust_sample:
        cs["READNUM"] = readnum[cs["LABEL"]]
    return clust_sample


def add_percent(clust_sample):
    n_sample = len(clust_sample)
    percent = defaultdict(int)
    for cs in clust_sample:
        percent[cs["LABEL"]] += 1 / n_sample
    percent = {key: round(val * 100, 3) for key, val in percent.items()}
    for cs in clust_sample:
        cs["PERCENT"] = percent[cs["LABEL"]]
    return clust_sample


def update_labels(clust_sample):
    """
    Allocate new labels according to the ranking by PERCENT
    """
    clust_sample.sort(key=lambda x: (-x["PERCENT"], x["LABEL"]))
    new_label = 1
    prev_label = clust_sample[0]["LABEL"]
    for cs in clust_sample:
        if prev_label != cs["LABEL"]:
            new_label += 1
        prev_label = cs["LABEL"]
        cs["LABEL"] = new_label
    return clust_sample



from pathlib import Path
import midsv
from collections import defaultdict


def mask_control(TEMPDIR: Path, DICT_ALLELE: dict, CONTROL_NAME: str):
    for allele in DICT_ALLELE.keys():
        if allele == "control":
            continue
        masks_control = defaultdict(list)
        path_midsv = Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_{allele}.jsonl")
        cssplit_control = midsv.read_jsonl(path_midsv)
        cssplit_control = [cs["CSSPLIT"].split(",") for cs in cssplit_control]
        coverage = len(cssplit_control)
        cssplit_control_transposed = list(zip(*cssplit_control))
        masks_loci = [False] * (len(cssplit_control_transposed) - 1)
        for i, cs in enumerate(cssplit_control_transposed):
            count_match = sum(1 for c in cs if c.startswith("=") or c == "N")
            percent_match = count_match / coverage * 100
            if percent_match < 1:
                masks_loci[i] = True
        masks_control[allele] = masks_loci
    return masks_control



from __future__ import annotations
from bisect import bisect_left
from collections import defaultdict
from itertools import groupby


def merge_dict1(d1, d2):
    d_merged = d1.copy()
    for k, v in d2.items():
        if k not in d_merged:
            d_merged[k] = v
    return d_merged


def join_listdicts(list_dict1: list[dict], list_dict2: list[dict], key: str) -> list[dict]:
    list_dict2.sort(key=lambda x: x[key])
    key_dict2 = [d[key] for d in list_dict2]
    dict_merged = []
    for d1 in list_dict1:
        key_dict1 = d1[key]
        if key_dict1 not in key_dict2:
            continue
        idx = bisect_left(key_dict2, key_dict1)
        d2 = list_dict2[idx]
        dict_merged.append(merge_dict1(d1, d2))
    return dict_merged


def transpose(cssplit: list[str]) -> list[tuple[str]]:
    cs_transposed = []
    for cs in cssplit:
        cs_transposed.append(cs["CSSPLIT"].split(","))
    return list(zip(*cs_transposed))


def call_percentage(cssplit_sample: list[str], diffloci: list[int]) -> list[dict]:
    """
    Call position weight matrix in defferent loci.
    Non defferent loci are annotated to "Match" or "Unknown(N)"
    """
    cssplit_sample_transposed = transpose(cssplit_sample)
    readnum_sample = len(cssplit_sample)
    con_percentage = []
    for idx, cs_transposed in enumerate(cssplit_sample_transposed):
        if idx in diffloci:
            count_cs = defaultdict(int)
            for cs in cs_transposed:
                count_cs[cs] += 1 / readnum_sample
            count_cs_sorted = dict(sorted(count_cs.items(), key=lambda x: x[1], reverse=True))
            con_percentage.append(count_cs_sorted)
        else:
            cons_cs = {"N": 1.0}
            for cs in cs_transposed:
                if cs.startswith("="):
                    cons_cs = {cs: 1.0}
                    break
            con_percentage.append(cons_cs)
    return con_percentage


# def call_percentage(cssplit_sample: list[str], cssplit_control: list[str]) -> list[dict]:
#     cssplit_control_transposed = transpose(cssplit_control)
#     cssplit_sample_transposed = transpose(cssplit_sample)
#     readnum_sample = len(cssplit_sample)
#     readnum_control = len(cssplit_control)
#     con_percentage = []
#     for i in range(len(cssplit_sample_transposed)):
#         d_control = defaultdict(int)
#         for cs in cssplit_control_transposed[i]:
#             d_control[cs] += 1 / readnum_control
#         d_sample = defaultdict(int)
#         for cs in cssplit_sample_transposed[i]:
#             d_sample[cs] += 1 / readnum_sample
#         d_sample_subtracted = defaultdict(int)
#         for key, value in d_sample.items():
#             if key.startswith("="):
#                 d_sample_subtracted[key] = value
#                 continue
#             value = max(0, value - d_control[key])
#             if value > 0:
#                 d_sample_subtracted[key] = value
#         d_sum = sum(d_sample_subtracted.values())
#         d = {key: value * 1 / d_sum for key, value in d_sample_subtracted.items()}
#         d = dict(sorted(d.items(), key=lambda x: x[1], reverse=True))
#         con_percentage.append(d)
#     return con_percentage


def call_sequence(cons_percentage_by_key: list[dict]) -> str:
    consensus_sequence = []
    for cons_per in cons_percentage_by_key:
        cons = max(cons_per, key=cons_per.get)
        if cons.startswith("="):
            cons = cons.replace("=", "")
        elif cons.startswith("-"):
            continue
        elif cons.startswith("*"):
            cons = cons[-1]
        elif cons.startswith("+"):
            cons_ins = cons.split("|")
            if cons_ins[-1].startswith("="):
                cons = cons.replace("=", "")
            elif cons_ins[-1].startswith("-"):
                cons = "".join(cons_ins[:-1])
            elif cons_ins[-1].startswith("*"):
                cons = "".join([*cons_ins[:-1], cons_ins[-1][-1]])
            cons = cons.replace("+", "")
            cons = cons.replace("|", "")
        consensus_sequence.append(cons)
    return "".join(consensus_sequence)


def call_allele_name(keys: tuple, cons_seq: str, DICT_ALLELE: dict) -> str:
    ALLELE, SV, LABEL, PERCENT = keys
    allele_name = f"allele{LABEL}_{ALLELE}"
    if SV:
        allele_name += "_sv"
    elif cons_seq == DICT_ALLELE[ALLELE]:
        allele_name += "_intact"
    else:
        allele_name += "_mutated"
    allele_name += f"_{PERCENT}%"
    return allele_name


def call(clust_sample: list[dict], diffloci_by_alleles: defaultdict[int], DICT_ALLELE: dict):
    result_sample = []
    cons_percentage = defaultdict(list)
    cons_sequence = defaultdict(list)
    clust_sample.sort(key=lambda x: x["LABEL"])
    for _, group in groupby(clust_sample, key=lambda x: x["LABEL"]):
        cssplit_sample = list(group)
        keys = (
            cssplit_sample[0]["ALLELE"],
            cssplit_sample[0]["SV"],
            cssplit_sample[0]["LABEL"],
            cssplit_sample[0]["PERCENT"],
        )
        diffloci = diffloci_by_alleles[f"{keys[0]}-{keys[1]}"]
        cons_per = call_percentage(cssplit_sample, diffloci)
        cons_seq = call_sequence(cons_per)
        allele_name = call_allele_name(keys, cons_seq, DICT_ALLELE)
        cons_percentage[allele_name] = cons_per
        cons_sequence[allele_name] = cons_seq
        for cs in cssplit_sample:
            cs["NAME"] = allele_name
        result_sample.extend(cssplit_sample)
    result_sample.sort(key=lambda x: x["LABEL"])
    return result_sample, cons_percentage, cons_sequence


# def call(TEMPDIR, clust_sample, DICT_ALLELE, SAMPLE_NAME, CONTROL_NAME):
#     result_sample = []
#     cons_percentage = defaultdict(list)
#     cons_sequence = defaultdict(list)
#     clust_sample.sort(key=lambda x: x["LABEL"])
#     for _, group in groupby(clust_sample, key=lambda x: x["LABEL"]):
#         clust = list(group)
#         keys = (clust[0]["ALLELE"], clust[0]["SV"], clust[0]["LABEL"], clust[0]["PERCENT"])
#         allele = keys[0]
#         cssplit_control = midsv.read_jsonl(Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_{allele}.jsonl"))
#         cssplit_sample = midsv.read_jsonl(Path(TEMPDIR, "midsv", f"{SAMPLE_NAME}_{allele}.jsonl"))
#         cssplit_sample = join_listdicts(clust, cssplit_sample, key="QNAME")
#         cons_per = call_percentage(cssplit_sample, cssplit_control)
#         cons_seq = call_sequence(cons_per)
#         allele_name = call_allele_name(keys, cons_seq, DICT_ALLELE)
#         cons_percentage[allele_name] = cons_per
#         cons_sequence[allele_name] = cons_seq
#         for cs in cssplit_sample:
#             cs["NAME"] = allele_name
#             del cs["RNAME"]
#             del cs["CSSPLIT"]
#         result_sample.extend(cssplit_sample)
#     result_sample.sort(key=lambda x: x["LABEL"])
#     return result_sample, cons_percentage, cons_sequence




from __future__ import annotations

from pathlib import Path
from urllib.request import urlopen
import re
import mappy
import midsv

########################################################################
# Make directories
########################################################################


def make_directories(TEMPDIR: Path):
    subdirectoris = ["cache", "fasta", "sam", "midsv", "report", "result"]
    for subdir in subdirectoris:
        Path(TEMPDIR, subdir).mkdir(parents=True, exist_ok=True)
    reportdirectories = ["HTML", "FASTA", "VCF", "BAM", ".igvjs"]
    for reportdir in reportdirectories:
        Path(TEMPDIR, "report", reportdir).mkdir(parents=True, exist_ok=True)


########################################################################
# Extract basename
########################################################################


def extract_basename(fastq_path: str) -> str:
    name = Path(fastq_path).name
    name = re.sub(r"\..*$", "", name)
    name = name.lstrip()
    if not name:
        raise AttributeError(f"{fastq_path} is not valid file name")
    else:
        return re.sub(r'[\\|/|:|?|.|,|\'|"|<|>|\| |]', "-", name)


########################################################################
# Convert allele file to dictionary type fasta format
########################################################################


def dictionize_allele(path_fasta: str) -> dict:
    header, sequence = [], []
    for name, seq, _ in mappy.fastx_read(path_fasta):
        name = name.lstrip()
        if not name:
            raise AttributeError(f"{path_fasta} contains an empty header")
        name = re.sub(r'[\\|/|:|?|.|,|\'|"|<|>|\| |]', "-", name)
        header.append(name)
        sequence.append(seq.upper())
    return {h: s for h, s in zip(header, sequence)}


########################################################################
# Fetch genome coodinate and chrom size
########################################################################


# TODO TEST
def fetch_coodinate(genome: str, ucsc_url: str, seq: str) -> dict:
    ucsc_blat = f"{ucsc_url}/cgi-bin/hgBlat?db={genome}&type=BLAT&userSeq={seq}"
    request = urlopen(ucsc_blat).read().decode("utf8").split("\n")
    coodinate = [x for x in request if x.count("100.0%")]
    if not coodinate:
        raise AttributeError(f"{seq} is not found in {genome}")
    else:
        coodinate = coodinate[0].split()
        return {"chr": coodinate[-5], "start": int(coodinate[-3]), "end": int(coodinate[-2]), "strand": coodinate[-4]}


def fetch_chrom_size(chrom: str, genome: str, goldenpath_url: str) -> int:
    url = f"{goldenpath_url}/{genome}/bigZips/{genome}.chrom.sizes"
    request = urlopen(url).read().decode("utf8").split("\n")
    for req in request:
        req = req.split("\t")
        if chrom == req[0]:
            chrom_size = int(req[1])
    return chrom_size


def get_coodinates_and_chromsize(
    TEMPDIR, GENOME, DICT_ALLELE, UCSC_URL, GOLDENPATH_URL, IS_CACHE_GENOME: bool
) -> tuple:
    path_genome_coodinates = Path(TEMPDIR, "cache", "genome_coodinates.jsonl")
    path_chrome_size = Path(TEMPDIR, "cache", "chrome_size.txt")
    if IS_CACHE_GENOME:
        GENOME_COODINATES = midsv.read_jsonl(path_genome_coodinates)[0]
        CHROME_SIZE = int(path_chrome_size.read_text())
    else:
        GENOME_COODINATES = fetch_coodinate(GENOME, UCSC_URL, DICT_ALLELE["control"])
        CHROME_SIZE = fetch_chrom_size(GENOME_COODINATES["chr"], GENOME, GOLDENPATH_URL)
    return GENOME_COODINATES, CHROME_SIZE


def cache_coodinates_and_chromsize(TEMPDIR, GENOME, GENOME_COODINATES, CHROME_SIZE):
    # Save info to the cache directory
    Path(TEMPDIR, "cache", "genome_symbol.txt").write_text(GENOME + "\n")
    midsv.write_jsonl([GENOME_COODINATES], Path(TEMPDIR, "cache", "genome_coodinates.jsonl"))
    Path(TEMPDIR, "cache", "chrome_size.txt").write_text(str(CHROME_SIZE))
    # Save info to the .igvjs directory
    Path(TEMPDIR, "report", ".igvjs", "genome_symbol.txt").write_text(GENOME + "\n")
    midsv.write_jsonl([GENOME_COODINATES], Path(TEMPDIR, "report", ".igvjs", "genome_coodinates.jsonl"))



import numpy as np
import scipy.stats as st
from sklearn.cluster import Birch
from scipy.sparse import csr_matrix
from collections import defaultdict


def predict_labels(xflatten, n_clusters=1):
    brc = Birch(n_clusters=n_clusters)
    brc.fit(xflatten)
    return brc.predict(xflatten)


def make_table(labels):
    count = defaultdict(int)
    for label in labels:
        count[label] += 1
    table = list(count.values())
    table.sort(reverse=True)
    return table


def chistatistic(prev_table, current_table, threshold=0.05) -> float:
    chisq_value, _, ddof, _ = st.chi2_contingency([prev_table, current_table])
    delta = sum(prev_table + current_table) * threshold ** 2
    pval = 1 - st.ncx2.cdf(chisq_value, ddof, delta)
    return pval


def return_labels(scores):
    # flatten from 3d to 2d
    xarr = np.array(scores)
    xflatten = np.reshape(xarr, (len(xarr), -1))
    xflatten = csr_matrix(xflatten)
    n_clusters = 1
    while True or n_clusters < 1000:
        prev_labels = predict_labels(xflatten, n_clusters=n_clusters)
        prev_table = make_table(prev_labels)
        current_labels = predict_labels(xflatten, n_clusters=n_clusters + 1)
        current_table = make_table(current_labels)
        if n_clusters == 1:
            prev_table += [0]
        else:
            current_table = current_table[:-1]
        pval = chistatistic(prev_table, current_table, threshold=0.05)
        if pval > 0.05:
            labels = prev_labels
            break
        n_clusters += 1
    return labels




from __future__ import annotations
from pathlib import Path
import re
import mappy
import hashlib
from urllib.request import urlopen
from urllib.error import URLError


def exists(input_file: str):
    if not Path(input_file).exists():
        raise FileNotFoundError(f"{input_file} is not found")


########################################################################
# Check if the sample is in the proper format.
########################################################################


def fastq_extension(fastq_path: str):
    if not re.search(r".fastq$|.fastq.gz$|.fq$|.fq.gz$", fastq_path):
        raise AttributeError(f"{fastq_path} requires extensions either 'fastq', 'fastq.gz', 'fq' or 'fq.gz'")


# File content
def fastq_content(fastq_path: str):
    name, seq, qual = [], [], []
    for n, s, q in mappy.fastx_read(fastq_path):
        name.append(n)
        seq.append(s)
        qual.append(q)
    if not (len(name) == len(seq) == len(qual) > 0):
        raise AttributeError(f"{fastq_path} is not a FASTQ format")


def fasta_content(fasta_path: str):
    name, seq = [], []
    for n, s, _ in mappy.fastx_read(fasta_path):
        name.append(n)
        seq.append(s)
    if not len(name) == len(seq) > 0:
        raise AttributeError(f"{fasta_path} is not a FASTA format")
    if len(name) > len(set(name)):
        raise AttributeError(f"{fasta_path} must include unique identifiers")
    if len(seq) > len(set(seq)):
        raise AttributeError(f"{fasta_path} must include unique DNA sequences")
    if name.count("control") == 0:
        raise AttributeError(f"{fasta_path} must include a 'control' sequence")


def check_files(SAMPLE: str, CONTROL: str, ALLELE: str) -> None:
    exists(CONTROL)
    exists(SAMPLE)
    exists(ALLELE)
    fastq_extension(CONTROL)
    fastq_content(CONTROL)
    fastq_extension(SAMPLE)
    fastq_content(SAMPLE)
    fasta_content(ALLELE)


########################################################################
# Check Cache
########################################################################


def is_cache_control(CONTROL: str, TEMPDIR: Path) -> bool:
    PATH_CACHE_HASH = Path(TEMPDIR, "cache", "control_hash.txt")
    IS_CACHE_CONTROL = False
    if PATH_CACHE_HASH.exists():
        current_hash = hashlib.sha256(Path(CONTROL).read_bytes()).hexdigest()
        cashed_hash = PATH_CACHE_HASH.read_text()
        if current_hash == cashed_hash:
            IS_CACHE_CONTROL = True
    return IS_CACHE_CONTROL


def is_cache_genome(GENOME: str, TEMPDIR: Path, IS_CACHE_CONTROL: bool) -> bool:
    PATH_CACHE_GENOME = Path(TEMPDIR, "cache", "genome_symbol.txt")
    IS_CACHE_GENOME = False
    if GENOME and IS_CACHE_CONTROL and PATH_CACHE_GENOME.exists():
        cashed_genome = PATH_CACHE_GENOME.read_text()
        if GENOME == cashed_genome:
            IS_CACHE_GENOME = True
    return IS_CACHE_GENOME


########################################################################
# Check genome and UCSC server
########################################################################


def available_url(urls: list[str]) -> tuple[str, bool]:
    flag_fail = False
    url = ""
    for url in urls:
        try:
            _ = urlopen(url)
        except URLError:
            flag_fail = True
        else:
            flag_fail = False
            break
    return url, flag_fail


def available_genome(genome: str, ucsc_url: str):
    url = f"{ucsc_url}/cgi-bin/das/{genome}/dna?segment=1:1,10"
    response = urlopen(url)
    content = response.read()
    response.close()
    if not content:
        raise AttributeError(
            f"{genome} is not listed in UCSC genome browser. Available genomes are in {ucsc_url}/cgi-bin/das/dsn"
        )


def check_and_fetch_genome(GENOME: str) -> tuple(str, str):
    # Check UCSC Server
    UCSC_URLS = [
        "https://genome.ucsc.edu/",
        "https://genome-asia.ucsc.edu/",
        "https://genome-euro.ucsc.edu/",
    ]
    UCSC_URL, flag_fail = available_url(UCSC_URLS)
    if flag_fail:
        raise URLError("UCSC Servers are currently down")
    # Check UCSC Download Server
    GOLDENPATH_URLS = [
        "https://hgdownload.cse.ucsc.edu/goldenPath",
        "http://hgdownload-euro.soe.ucsc.edu/goldenPath",
    ]
    GOLDENPATH_URL, flag_fail = available_url(GOLDENPATH_URLS)
    if flag_fail:
        raise URLError("UCSC Download Servers are currently down")
    # Check input genome
    available_genome(GENOME, UCSC_URL)
    return UCSC_URL, GOLDENPATH_URL



import os
import re
from pathlib import Path
import http.server
import socketserver
import webbrowser
import socket
from contextlib import closing
from jinja2 import Template, Environment, FileSystemLoader


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def execute(name: str):
    DIR_IGVJS = Path("DAJINResults", name, ".igvjs")
    if not DIR_IGVJS.exists():
        raise FileNotFoundError("BAM files for DAJIN view is not found. Execute DAJIN first.")
    env = Environment(loader=FileSystemLoader("./", encoding="utf8"))
    template = env.get_template("src/DAJIN2/template_igvjs.html")
    params_genome = {"genome": {"exist": False}}
    params_reference = dict()
    path_genome = Path(DIR_IGVJS, "genome_symbol.txt")
    if path_genome.exists():
        path_coodinates = Path(DIR_IGVJS, "genome_coodinates.jsonl")
        GENOME = path_genome.read_text().strip()
        CHROME, START, END, _ = eval(path_coodinates.read_text().strip()).values()
        params_genome = {"genome": {"exist": True, "genome": GENOME, "locus": f"{CHROME}:{START}-{END}"}}
    else:
        params_reference = {"reference": {"urlfa": "control.fasta", "urlfai": "control.fasta" + ".fai"}}

    bamnames = []
    bamurls = []
    baiurls = []
    for bam in Path("DAJINResults", name, ".igvjs").iterdir():
        if not re.search(r"bam$", str(bam)):
            continue
        bamnames.append(bam.stem)
        bamurls.append(str(bam.name))
        baiurls.append(str(bam.name) + ".bai")
    bamnames.sort()
    bamurls.sort()
    baiurls.sort()
    contents = [{"samplename": n, "urlbam": b, "urlbai": i} for n, b, i in zip(bamnames, bamurls, baiurls)]

    params_tracks = """{
        "tracks":
            {{ contents }},
    }"""
    params_tracks = Template(params_tracks)
    params_tracks = eval(params_tracks.render(contents=contents))

    params_genome.update(**params_reference, **params_tracks)
    params = params_genome.copy()

    HTML_IGVJS = template.render(params)
    Path(DIR_IGVJS, "index.html").write_text(HTML_IGVJS)
    PORT = find_free_port()
    Handler = http.server.SimpleHTTPRequestHandler
    os.chdir(Path("DAJINResults", name, ".igvjs"))
    with socketserver.TCPServer(("", PORT), Handler) as httpd:
        print(f"serving at port: http://127.0.0.1:{PORT}")
        httpd.serve_forever()
        webbrowser.open(f"http://127.0.0.1:{PORT}", autoraise=True)
    os.chdir("../../../")



from __future__ import annotations
import cstag
import textwrap


def to_fasta(header: str, cons_seq: str) -> str:
    header = ">" + header
    cons_seq_wrap = textwrap.wrap(cons_seq, 80)
    fasta = "\n".join([header, *cons_seq_wrap]) + "\n"
    return fasta


def to_html(header: str, cons_per: list[dict]) -> str:
    cons_cssplit = [max(cons, key=cons.get) for cons in cons_per]
    prev_cstag = cons_cssplit[0]
    cons_cstag = []
    cons_cstag.append(prev_cstag)
    for current_cstag in cons_cssplit[1:]:
        if "=" == prev_cstag[0] == current_cstag[0]:
            cons_cstag.append(current_cstag[1])
        elif "-" == prev_cstag[0] == current_cstag[0]:
            cons_cstag.append(current_cstag[1])
        elif "+" == current_cstag[0]:
            cs_ins = current_cstag.replace("+", "").split("|")
            ins = "+" + "".join(cs_ins[:-1]) + cs_ins[-1]
            cons_cstag.append(ins)
        else:
            cons_cstag.append(current_cstag)
        prev_cstag = current_cstag
    cons_cstag = "".join(cons_cstag)
    return cstag.to_html(cons_cstag, header)


def to_vcf(header: str, cons_per: list[dict]) -> str:
    pass




from flask import Flask, render_template, request
from werkzeug.utils import secure_filename
import webbrowser
from threading import Timer
from waitress import serve
from pathlib import Path
import pandas as pd
from . import batch


def open_browser():
    webbrowser.open_new("http://127.0.0.1:2000/")


def upload_files(files):
    for file in files:
        path = Path(app.config["UPLOAD_FOLDER"], secure_filename(file.filename))
        file.save(path)


def get_path_of_uploaded_files(UPLOAD_FOLDER, files):
    paths = []
    for file in files:
        path = str(UPLOAD_FOLDER) + "/" + secure_filename(file.filename)
        paths.append(path)
    return paths


app = Flask(__name__)


@app.route("/")
def root_page():
    return render_template("index.html")


@app.route("/submit", methods=["POST"])
def submit():
    name = request.form.get("name")
    UPLOAD_FOLDER = Path("DAJINResults", ".tempdir", name, "upload")
    UPLOAD_FOLDER.mkdir(parents=True, exist_ok=True)
    app.config["UPLOAD_FOLDER"] = UPLOAD_FOLDER

    files = request.files.getlist("sample")
    upload_files(files)
    PATH_SAMPLE = get_path_of_uploaded_files(UPLOAD_FOLDER, files)

    files = request.files.getlist("control")
    upload_files(files)
    PATH_CONTROL = get_path_of_uploaded_files(UPLOAD_FOLDER, files)

    files = request.files.getlist("allele")
    upload_files(files)
    PATH_ALLELE = get_path_of_uploaded_files(UPLOAD_FOLDER, files)

    genome = request.form.get("genome")
    threads = request.form.get("threads")
    if threads is None:
        threads = 1

    df = pd.DataFrame({"sample": PATH_SAMPLE})
    df["name"] = name
    df["control"] = PATH_CONTROL[0]
    df["allele"] = PATH_ALLELE[0]
    if genome:
        df["genome"] = genome
    df.to_csv(Path("DAJINResults", ".tempdir", name, "upload", "batch.csv"), index=False)

    arguments = dict()
    arguments["file"] = str(Path("DAJINResults", ".tempdir", name, "upload", "batch.csv"))
    arguments["threads"] = threads
    arguments["debug"] = False
    batch.execute(arguments)

    return f"""
    name={name}
    sample={PATH_SAMPLE}
    control={PATH_CONTROL}
    allele={PATH_ALLELE}
    genome={genome}
    threads={threads}
    arguments={arguments["file"]}
    """


def execute():
    print("Assess 'http://127.0.0.1:2000/' if browser does not automatically open.")
    Timer(1, open_browser).start()
    serve(app, host="0.0.0.0", port=2000)




import re


def detect_sv(CSSPLIT: str, threshold: int = 50) -> bool:
    is_SV = False
    cssplit = CSSPLIT.replace(",", "")
    if "N" * threshold in cssplit:
        is_SV = True
    if re.search(rf"(\+[ACGTN]\|){{{threshold}}}", cssplit):
        is_SV = True
    if re.search(rf"(\-[ACGTN]){{{threshold}}}", cssplit):
        is_SV = True
    if re.search(rf"(\*[ACGTN][ACGTN]){{{threshold}}}", cssplit):
        is_SV = True
    if re.search(r"[acgtn]", cssplit):
        is_SV = True
    return is_SV



import midsv
from pathlib import Path


def output_midsv(TEMPDIR, path_sam, DICT_ALLELE):
    name_fasta = path_sam.stem.split("_")[1]
    if name_fasta not in set(DICT_ALLELE.keys()):
        return
    sam = midsv.read_sam(path_sam)
    midsv_jsonl = midsv.transform(sam, midsv=False, cssplit=True, qscore=False)
    output_jsonl = Path(TEMPDIR, "midsv", f"{path_sam.stem}.jsonl")
    midsv.write_jsonl(midsv_jsonl, output_jsonl)



from __future__ import annotations
from pathlib import Path
from itertools import groupby
import pandas as pd
import wslPath
from .core import core
from .postprocess import report


def execute(arguments: dict[str]):
    path_batchfile = arguments["file"]
    threads = arguments["threads"]
    # debug = arguments["debug"]

    ###############################################################################
    # Check arguments
    ###############################################################################
    try:
        path_batchfile = wslPath.toPosix(path_batchfile)
    except ValueError:
        pass

    # ----------------------------------------------------------
    # Check file exists
    # ----------------------------------------------------------
    if not Path(path_batchfile).exists():
        raise FileNotFoundError(f"'{path_batchfile}' does not exist.")

    try:
        df_batchfile = pd.read_excel(path_batchfile)
        inputs = []
        inputs = df_batchfile.columns.to_list()
        inputs += df_batchfile.values.tolist()
    except ValueError:
        inputs = [s.split(",") for s in Path(path_batchfile).read_text().strip().split("\n")]

    # ----------------------------------------------------------
    # Check Column
    # ----------------------------------------------------------
    keys = inputs[0]
    keys_set = set(keys)
    keys_required = set(["sample", "control", "allele", "name"])
    keys_all = set(["sample", "control", "allele", "name", "genome"])

    if not keys_required.issubset(keys_set):
        raise ValueError(f"{path_batchfile} must contain 'sample', 'control', and 'allele' in the header")

    if not keys_set.issubset(keys_all):
        raise ValueError(
            f"Accepted header names of {path_batchfile} are 'sample', 'control', 'allele', 'name', and 'genome'."
        )

    ##############################################################################
    # Perform DAJIN
    ##############################################################################
    index_name = [i for i, key in enumerate(keys) if key == "name"][0]
    contents = inputs[1:]
    contents.sort(key=lambda x: x[index_name])

    for name, groups in groupby(contents, key=lambda x: x[index_name]):
        for group in groups:
            args = {h: g for h, g in zip(keys, group)}
            args["threads"] = threads
            core.execute(args)
        report.report(name)



import re
import numpy as np


def extract_cssplit_at_diffloci(cssplit_sample, diffloci):
    cssplit_diffloci = []
    for cssplit in cssplit_sample:
        cssplit = cssplit.split(",")
        diff_cs = []
        for locus in diffloci:
            diff_cs.append(cssplit[locus])
        cssplit_diffloci.append(diff_cs)
    return cssplit_diffloci


# Summation scores at each loci
def sum_scores(cssplit_diffloci):
    length = len(cssplit_diffloci[0])
    scores = [[0 for _ in range(7)] for _ in range(length)]
    for i, cssplit in enumerate(zip(*cssplit_diffloci)):
        for cs in cssplit:
            if cs.startswith("-"):
                scores[i][4] += 1
            elif cs.startswith("*"):
                scores[i][5] += 1
            elif cs.startswith("N"):
                scores[i][6] += 1
            elif cs.startswith("+"):
                cs_insertion = cs.split("|")
                if cs_insertion[-1].startswith("="):
                    scores[i][0] += 1
                elif cs_insertion[-1].startswith("-"):
                    scores[i][1] += 1
                elif cs_insertion[-1].startswith("*"):
                    scores[i][2] += 1
                elif cs_insertion[-1].startswith("N"):
                    scores[i][3] += 1
    return scores


def score_repeats(i, cssplit, scores, length, operant, inversion=False):
    j = i
    while j < length and cssplit[j].startswith(operant):
        j += 1
    score_repeat = j - i
    if inversion:
        score_repeat *= -1
    for idx in range(i, j):
        if operant == "-":
            scores[idx][4] = score_repeat
        if operant == "*":
            scores[idx][5] = score_repeat
        if operant == "N":
            scores[idx][6] = score_repeat
    return j - 1, scores


def score_repeats_insertion(i, cs, scores, inversion=False):
    cs_ins = cs.split("|")
    score = len(cs_ins)
    if inversion:
        score *= -1
    if cs_ins[-1].startswith("="):
        scores[i][0] = score
    elif cs_ins[-1].startswith("-"):
        scores[i][1] = score
    elif cs_ins[-1].startswith("*"):
        scores[i][2] = score
    else:
        scores[i][3] = score
    return i, scores


def make_scores(cssplit_sample, diffloci):
    cssplit_diffloci = extract_cssplit_at_diffloci(cssplit_sample, diffloci)
    scores_summed = sum_scores(cssplit_diffloci)
    scores_summed = np.array(scores_summed)
    length = len(diffloci)
    scores_all = []
    for cssplit in cssplit_diffloci:
        # IM, ID, IS, IN, D, S, N
        scores = [[0 for _ in range(7)] for _ in range(length)]
        i = 0
        while i < length:
            cs = cssplit[i]
            inversion = False
            if re.search(r"[acgtn]", cs):
                inversion = True
            if cs.startswith("-"):
                i, scores = score_repeats(i, cssplit, scores, length, "-", inversion)
            elif cs.startswith("*"):
                i, scores = score_repeats(i, cssplit, scores, length, "*", inversion)
            elif cs.startswith("N"):
                i, scores = score_repeats(i, cssplit, scores, length, "N", inversion)
            elif cs.startswith("+"):
                i, scores = score_repeats_insertion(i, cs, scores, inversion)
            i += 1
        scores = np.array(scores)
        scores = (scores * scores_summed).tolist()
        scores_all.append(scores)
    return scores_all



from __future__ import annotations
import midsv
from pathlib import Path
from itertools import groupby


def calc_match(CSSPLIT: str) -> float:
    match_score = CSSPLIT.count("=")
    match_score -= CSSPLIT.count("+")  # insertion
    match_score -= sum(cs.islower() for cs in CSSPLIT)  # inversion
    return match_score / len(CSSPLIT.split(","))


def classify_alleles(path_midsv: Path, sample_name: str) -> list[dict]:
    score_of_each_allels = []
    for midsvpath in path_midsv:
        ALLELE = midsvpath.stem.replace(f"{sample_name}_", "")
        for midsvcsv in midsv.read_jsonl(midsvpath):
            QNAME = midsvcsv["QNAME"]
            CSSPLIT = midsvcsv["CSSPLIT"]
            SCORE = calc_match(CSSPLIT)
            score_of_each_allels.append({"QNAME": QNAME, "ALLELE": ALLELE, "SCORE": SCORE, "CSSPLIT": CSSPLIT})
    score_of_each_allels.sort(key=lambda x: x["QNAME"])
    possible_allele = []
    for _, group in groupby(score_of_each_allels, key=lambda x: x["QNAME"]):
        max_score = -float("inf")
        for readinfo in group:
            if readinfo["SCORE"] > max_score:
                max_score = readinfo["SCORE"]
                max_read = readinfo
                del max_read["SCORE"]
        possible_allele.append(max_read)
    return possible_allele



from __future__ import annotations
import re
import scipy.stats as st


def replaceNtoMatch(sample_cs: list[str]) -> list[str]:
    # Replace N to @ at the left ends
    for i, cs in enumerate(sample_cs):
        if cs != "N":
            break
        sample_cs[i] = "="
    # Replace N to @ at the right ends
    sample_cs = sample_cs[::-1]
    for i, cs in enumerate(sample_cs):
        if cs != "N":
            break
        sample_cs[i] = "="
    sample_cs = sample_cs[::-1]
    return sample_cs


def make_table(cssplit_sample: list[str], cssplit_control: list[str]) -> tuple(list, list):
    # Create empty templates
    length = len(cssplit_control[0].split(","))
    # table of Deletion, Others
    table_sample = [[1, 1] for _ in range(length)]
    table_control = [[1, 1] for _ in range(length)]
    # Calculate match and mismatch
    for sample_cs, control_cs in zip(cssplit_sample, cssplit_control):
        sample_cs = sample_cs.split(",")
        sample_cs = replaceNtoMatch(sample_cs)
        control_cs = control_cs.split(",")
        control_cs = replaceNtoMatch(control_cs)
        for i, cs in enumerate(sample_cs):
            if cs.startswith("="):
                continue
            if cs.startswith("-"):
                table_sample[i][0] += 1
            elif cs.startswith("+"):
                table_sample[i][1] += cs.count("+")
            else:
                table_sample[i][1] += 1
        for i, cs in enumerate(control_cs):
            if cs.startswith("="):
                continue
            if cs.startswith("-"):
                table_control[i][0] += 1
            elif cs.startswith("+"):
                table_control[i][1] += cs.count("+")
            else:
                table_control[i][1] += 1
    return table_sample, table_control


def mask_table_control(table_control, masks_control):
    for i, is_mask in enumerate(masks_control):
        if is_mask:
            table_control[i] = [1, 1]
    return table_control


def chistatistic(s_table, c_table, threshold=0.05) -> float:
    chisq_value, _, ddof, _ = st.chi2_contingency([s_table, c_table])
    delta = sum(s_table + c_table) * threshold ** 2
    pval = 1 - st.ncx2.cdf(chisq_value, ddof, delta)
    return pval


def screen_different_loci(
    cssplit_sample: list[str],
    cssplit_control: list[str],
    sequence: str,
    masks_control: list[bool],
    alpha: float = 0.01,
    threshold: float = 0.05,
) -> list[dict]:
    """Scoring mutation (insertion, deletion, substitution, inversion, and unknow)
    at significant loci between sample and control

    Args:
        cssplit_sample (list[str]): List of sample's CSSPLITs
        cssplit_control (list[str]): List of control's CSSPLITs

    Returns:
        list[dict]: Scores at significantly different base loci
    """
    different_loci = []
    coverage = min(len(cssplit_sample), len(cssplit_control))
    table_sample, table_control = make_table(cssplit_sample, cssplit_control)
    table_control = mask_table_control(table_control, masks_control)
    repeat_regrex = r"A{4,}|C{4,}|G{4,}|T{4,}|N{4,}"
    repeat_span = (x.span() for x in re.finditer(repeat_regrex, sequence))
    try:
        repeat_start, repeat_end = next(repeat_span)
        repeat_end -= 1
    except StopIteration:
        # Compare each nucleotide locus when sequence does not have a repeat
        for i, (s, c) in enumerate(zip(table_sample, table_control)):
            pval = chistatistic([coverage, sum(s)], [coverage, sum(c)], threshold)
            if pval < alpha:
                different_loci.append(i)
        return different_loci
    # Sum of deletion and other mismatchs
    repeat_del = [1, 1]
    repeat_other = [1, 1]
    for i, (s, c) in enumerate(zip(table_sample, table_control)):
        if repeat_start <= i < repeat_end:
            repeat_del[0] += s[0]
            repeat_del[1] += c[0]
            repeat_other[0] += s[1]
            repeat_other[1] += c[1]
        elif i == repeat_end:
            repeat_del[0] += s[0]
            repeat_del[1] += c[0]
            repeat_other[0] += s[1]
            repeat_other[1] += c[1]
            # statistics to deletion
            pval = chistatistic([coverage, repeat_del[0]], [coverage, repeat_del[1]], threshold)
            if pval < pow(alpha, 3):
                for j in range(repeat_start, repeat_end + 1):
                    s, c = table_sample[j], table_control[j]
                    pval = chistatistic([coverage, s[0]], [coverage, c[0]], threshold)
                    if pval < pow(alpha, 3):
                        different_loci.append(j)
            # statistics to others
            pval = chistatistic([coverage, repeat_other[0]], [coverage, repeat_other[1]], threshold)
            if pval < alpha:
                for j in range(repeat_start, repeat_end + 1):
                    s, c = table_sample[j], table_control[j]
                    pval = chistatistic([coverage, s[1]], [coverage, c[1]], threshold)
                    if pval < alpha and j not in different_loci:
                        different_loci.append(j)
            repeat_del = [1, 1]
            repeat_other = [1, 1]
            try:
                repeat_start, repeat_end = next(repeat_span)
                repeat_end -= 1
            except StopIteration:
                pass
        else:
            pval = chistatistic([coverage, sum(s)], [coverage, sum(c)], threshold)
            if pval < alpha:
                different_loci.append(i)
    return different_loci



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt") as requirements_file:
    install_requirements = requirements_file.read().splitlines()

setuptools.setup(
    name="DAJIN2",
    version="0.1.3",
    author="Akihiro Kuno",
    author_email="akuno@md.tsukuba.ac.jp",
    description="One-step genotyping tools for Nanopore amplicon sequencing",
    long_description=long_description,
    url="https://github.com/akikuno/DAJIN2",
    install_requires=install_requirements,
    packages=setuptools.find_packages(where="src",),
    package_dir={"": "src"},
    entry_points={"console_scripts": ["DAJIN2=DAJIN2.DAJIN2:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
    ],
    python_requires=">=3.7",
)



import os
import argparse

from . import single
from . import batch
from . import gui
from . import view


def update_threads(threads):
    os_cpus = int(os.cpu_count())
    new_threads = max(1, threads)
    if threads > os_cpus:
        new_threads = os_cpus - 1
    return new_threads


def main():
    parser = argparse.ArgumentParser()

    ###############################################################################
    # Single mode
    ###############################################################################

    parser.add_argument("-s", "--sample", type=str, help="Give the full path to a control FASTQ file")
    parser.add_argument("-c", "--control", type=str, help="Give the full path to a control FASTQ file")
    parser.add_argument("-a", "--allele", type=str, help="Give the full path to allele FASTA file")
    parser.add_argument("-n", "--name", type=str, help="Output name of the report")
    parser.add_argument(
        "-g", "--genome", type=str, default="", help="Reference genome ID (e.g hg38, mm10) [default: '']"
    )
    parser.add_argument("-t", "--threads", type=int, default=1, help="Number of threads [default: 1]")
    parser.add_argument("-v", "--version", action="version", version="DAJIN2 version 1.0.0")
    parser.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)

    ###############################################################################
    # Batch mode
    ###############################################################################

    def batchmode(args):
        threads = update_threads(args.threads)
        arguments = dict()
        arguments["file"] = args.file
        arguments["threads"] = threads
        arguments["debug"] = args.debug
        batch.execute(arguments)

    subparser = parser.add_subparsers()
    parser_batch = subparser.add_parser("batch", help="DAIJN2 batch mode")
    parser_batch.add_argument("-f", "--file", required=True, type=str, help="The comma-separated batch file.")
    parser_batch.add_argument("-t", "--threads", default=1, type=int, help="Number of threads [default: 1]")
    parser_batch.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)
    parser_batch.set_defaults(handler=batchmode)

    ###############################################################################
    # GUI mode
    ###############################################################################

    def guimode(args):
        gui.execute()

    parser_gui = subparser.add_parser("gui", help="DAIJN2 GUI mode")
    parser_gui.set_defaults(handler=guimode)

    ###############################################################################
    # View mode
    ###############################################################################

    def viewmode(args):
        view.execute(args.name)

    parser_view = subparser.add_parser("view", help="DAIJN2 View mode to launch igvjs")
    parser_view.add_argument("-n", "--name", required=True, type=str, help="Output name of the report")
    parser_view.set_defaults(handler=viewmode)

    ###############################################################################
    # Parse arguments
    ###############################################################################

    args = parser.parse_args()

    if hasattr(args, "handler"):
        args.handler(args)
    else:
        if args.sample is None:
            raise AttributeError("the following arguments are required: -s/--sample")
        if args.control is None:
            raise AttributeError("the following arguments are required: -c/--control")
        if args.allele is None:
            raise AttributeError("the following arguments are required: -a/--allele")
        if args.name is None:
            raise AttributeError("the following arguments are required: -n/--name")
        threads = update_threads(args.threads)
        arguments = dict()
        arguments["sample"] = args.sample
        arguments["control"] = args.control
        arguments["allele"] = args.allele
        arguments["name"] = args.name
        if args.genome:
            arguments["genome"] = args.genome
        arguments["threads"] = threads
        arguments["debug"] = args.debug
        single.execute(arguments)


if __name__ == "__main__":
    main()



from __future__ import annotations
import re
import random
from pathlib import Path
from typing import Union
from collections import defaultdict
from itertools import groupby
import pysam
import midsv


def revcomp(sequence: str) -> str:
    complement = {"A": "T", "C": "G", "G": "C", "T": "A"}
    return "".join(complement[nt] for nt in sequence[::-1])


def split_cigar(CIGAR: str) -> list[str]:
    cigar = re.split(r"([MDISH=X])", CIGAR)
    n = len(cigar)
    cigar_split = []
    for i, j in zip(range(0, n, 2), range(1, n, 2)):
        cigar_split.append(cigar[i] + cigar[j])
    return cigar_split


def calc_length(CIGAR: str) -> int:
    cigar = split_cigar(CIGAR)
    seq_length = 0
    for c in cigar:
        if re.search(r"[MD=X]", c[-1]):
            seq_length += int(c[:-1])
    return seq_length


def write_sam(sam: list[list[str]], path_sam: Union[str, Path]):
    path_sam = str(path_sam)
    sam = ["\t".join(s) for s in sam]
    sam = "\n".join(sam)
    Path(path_sam).write_text(sam + "\n")


def remove_overlapped_reads(sam: list[list[str]]) -> list[list[str]]:
    sam_headers = [s for s in sam if s[0].startswith("@")]
    sam_contents = [s for s in sam if not s[0].startswith("@") and s[9] != "*"]
    sam_contents.sort(key=lambda x: [x[0], int(x[3])])
    sam_trimmed = sam_headers.copy()
    for _, group in groupby(sam_contents, key=lambda x: x[0]):
        alignments = list(group)
        if len(alignments) == 1:
            sam_trimmed.append(alignments[0])
            continue
        idx = 0
        flag_overlap = False
        while idx < len(alignments) - 1:
            prev_alignment = alignments[idx]
            next_alignment = alignments[idx + 1]
            prev_start = int(prev_alignment[3])
            prev_cigar = prev_alignment[5]
            prev_alignment_len = calc_length(prev_cigar)
            prev_end = prev_start + prev_alignment_len
            next_start = int(next_alignment[3])
            if prev_end >= next_start:
                flag_overlap = True
                break
            idx += 1
        if flag_overlap:
            continue
        else:
            sam_trimmed += alignments
    return sam_trimmed


def trim_softclip(CIGAR: str, SEQ: str) -> str:
    cigar_split = split_cigar(CIGAR)
    if cigar_split[0].endswith("S"):
        SEQ = SEQ[int(cigar_split[0][:-1]) :]
    if cigar_split[-1].endswith("S"):
        SEQ = SEQ[: -int(cigar_split[-1][:-1])]
    return SEQ


def remove_microhomology(sam: list[list[str]]) -> list[list[str]]:
    sam_headers = [s for s in sam if s[0].startswith("@")]
    sam_contents = [s for s in sam if not s[0].startswith("@") and s[9] != "*"]
    sam_contents.sort(key=lambda x: [x[0], int(x[3])])
    sam_trimmed = sam_headers.copy()
    for _, group in groupby(sam_contents, key=lambda x: x[0]):
        alignments = list(group)
        if len(alignments) == 1:
            sam_trimmed.append(alignments[0])
            continue
        idx = 0
        while idx < len(alignments) - 1:
            prev_align = alignments[idx]
            next_align = alignments[idx + 1]
            prev_cigar = prev_align[5]
            next_cigar = next_align[5]
            prev_seq = prev_align[9]
            next_seq = next_align[9]
            prev_qual = prev_align[10]
            next_qual = next_align[10]
            # trim softclips of sequence
            prev_seq_trimmed = trim_softclip(prev_cigar, prev_seq)
            next_seq_trimmed = trim_softclip(next_cigar, next_seq)
            # trim softclips of quality
            prev_qual_trimmed = trim_softclip(prev_cigar, prev_qual)
            next_qual_trimmed = trim_softclip(next_cigar, next_qual)
            if prev_seq_trimmed == next_seq_trimmed:
                sam_trimmed.append(prev_align)
                sam_trimmed.append(next_align)
                idx += 1
                continue
            i = 1
            len_microhomology = 0
            while i < min(len(prev_seq_trimmed), len(next_seq_trimmed)):
                if prev_seq_trimmed[-i:] == next_seq_trimmed[:i] and prev_qual_trimmed[-i:] == next_qual_trimmed[:i]:
                    len_microhomology = i
                i += 1
            if len_microhomology == 0:
                sam_trimmed.append(prev_align)
                sam_trimmed.append(next_align)
                idx += 1
                continue
            # ----------------------
            # format
            # ----------------------
            # start
            next_align[3] = str(int(next_align[3]) + len_microhomology)
            # cigar
            next_cigar_split = [c for c in split_cigar(next_cigar) if not re.search(r"[SH]$", c)]
            next_cigar_split[0] = str(int(next_cigar_split[0][:-1]) - len_microhomology) + next_cigar_split[0][-1]
            next_align[5] = "".join(next_cigar_split)
            # sequence
            next_align[9] = next_seq_trimmed[len_microhomology:]
            # quality
            next_align[10] = next_qual_trimmed[len_microhomology:]
            # ----------------------
            # finish
            # ----------------------
            sam_trimmed.append(prev_align)
            sam_trimmed.append(next_align)
            idx += 1
    return sam_trimmed


def reverse_sam(sam_contents: list[list[str]], genome_end: int) -> list[str]:
    sam_flags = [str(s) for s in [0, 16, 2048, 2064]]
    sam_reversed = []
    for sam_content in sam_contents:
        sam_update = sam_content.copy()
        sam_flag = sam_content[1]
        if sam_flag == sam_flags[0]:
            sam_flag = sam_flags[1]
        elif sam_flag == sam_flags[1]:
            sam_flag = sam_flags[0]
        elif sam_flag == sam_flags[2]:
            sam_flag = sam_flags[3]
        else:
            sam_flag = sam_flags[2]
        sam_update[1] = sam_flag
        sam_cigar = sam_content[5]
        sam_cigar = "".join(split_cigar(sam_cigar)[::-1])
        sam_update[5] = sam_cigar
        sam_start = int(sam_content[3])
        sam_length = calc_length(sam_cigar)
        sam_update[3] = str(genome_end - (sam_start + sam_length) + 2)
        sam_update[9] = revcomp(sam_content[9])
        sam_update[10] = sam_content[10][::-1]
        sam_reversed.append(sam_update)
    return sam_reversed


def realign(sam: list[list[str]], genome_coodinates: dict, chrom_size: int) -> list[str]:
    sam_headers = [s for s in sam if s[0].startswith("@")]
    sam_contents = [s for s in sam if not s[0].startswith("@")]
    for s in sam_headers:
        if s[0] != "@SQ":
            continue
        s[1] = f'SN:{genome_coodinates["chr"]}'
        s[2] = f"LN:{chrom_size}"
    for s in sam_contents:
        s[2] = genome_coodinates["chr"]
    if genome_coodinates["strand"] == "-":
        sam_contents = reverse_sam(sam_contents, genome_coodinates["end"])
    else:
        s[3] = str(int(s[3]) + genome_coodinates["start"] - 1)
    return sam_headers + sam_contents


def group_by_name(sam_contents: list[str], clust_sample: list[dict]) -> dict[list]:
    sam_contents.sort()
    clust_sample_qname = sorted(clust_sample, key=lambda x: x["QNAME"])
    clust_sample_qname_set = set()
    for qnames in clust_sample_qname:
        qname = qnames["QNAME"]
        clust_sample_qname_set.add(qname)
    sam_groups = defaultdict(list)
    idx_left = 0
    idx_right = 0
    while idx_left < len(sam_contents) and idx_right < len(clust_sample_qname):
        read_left = sam_contents[idx_left][:-1]
        read_right = clust_sample_qname[idx_right]
        qname_left = read_left[0]
        qname_right = read_right["QNAME"]
        if qname_left not in clust_sample_qname_set:
            idx_left += 1
            continue
        if qname_left == qname_right:
            key = read_right["NAME"]
            sam_groups[key].append(read_left)
            idx_left += 1
        else:
            idx_right += 1
    return sam_groups


###############################################################################
# igvjs
###############################################################################


def subset_qnames(RESULT_SAMPLE, readnum: int = 100) -> dict[set[str]]:
    qnames_by_name = defaultdict(set)
    for name, group in groupby(RESULT_SAMPLE, key=lambda x: x["NAME"]):
        group = list(group)
        qnames = [res["QNAME"] for res in group[:readnum]]
        qnames_by_name[name] = set(qnames)
    return qnames_by_name


def subset_reads(name, sam_content, qnames_by_name):
    qnames = qnames_by_name[name]
    sam_subset = [sam for sam in sam_content if sam[0] in qnames]
    return sam_subset


###############################################################################
# Output
###############################################################################


def output_bam(TEMPDIR, RESULT_SAMPLE, SAMPLE_NAME, GENOME, GENOME_COODINATES, CHROME_SIZE, THREADS):
    randomnum = random.randint(100_000, 999_999)
    path_sam = Path(TEMPDIR, "sam", f"{SAMPLE_NAME}_control.sam")
    sam = midsv.read_sam(path_sam)

    sam_update = sam.copy()
    sam_update = remove_overlapped_reads(sam_update)
    sam_update = remove_microhomology(sam_update)

    path_sam_update = Path(TEMPDIR, "report", "bam", f"tmp{randomnum}_{SAMPLE_NAME}_control.sam")
    if GENOME:
        sam_update = realign(sam_update, GENOME_COODINATES, CHROME_SIZE)
        write_sam(sam_update, path_sam_update)
    else:
        write_sam(sam_update, path_sam_update)

    path_bam = Path(TEMPDIR, "report", "bam", f"{SAMPLE_NAME}.bam")
    pysam.sort("-@", f"{THREADS}", "-o", str(path_bam), str(path_sam_update))
    pysam.index("-@", f"{THREADS}", str(path_bam))

    sam_headers = [s for s in sam_update if s[0].startswith("@")]
    sam_contents = [s for s in sam_update if not s[0].startswith("@")]
    sam_groups = group_by_name(sam_contents, RESULT_SAMPLE)
    qnames_by_name = subset_qnames(RESULT_SAMPLE)

    for name, sam_content in sam_groups.items():
        # BAM
        path_sam = Path(TEMPDIR, "report", "bam", f"tmp{randomnum}_{name}.sam")
        path_bam = Path(TEMPDIR, "report", "bam", f"{SAMPLE_NAME}_{name}.bam")
        write_sam(sam_headers + sam_content, path_sam)
        pysam.sort("-@", f"{THREADS}", "-o", str(path_bam), str(path_sam))
        pysam.index("-@", f"{THREADS}", str(path_bam))
        # igvjs
        sam_subset = subset_reads(name, sam_content, qnames_by_name)
        path_sam = Path(TEMPDIR, "report", "bam", f"tmp{randomnum}_{name}_subset.sam")
        path_bam = Path(TEMPDIR, "report", ".igvjs", f"{SAMPLE_NAME}_{name}.bam")
        write_sam(sam_headers + sam_subset, path_sam)
        pysam.sort("-@", f"{THREADS}", "-o", str(path_bam), str(path_sam))
        pysam.index("-@", f"{THREADS}", str(path_bam))

    # Remove temporary files
    sam_temp = Path(TEMPDIR, "report", "bam").glob(f"tmp{randomnum}*.sam")
    [s.unlink() for s in sam_temp]
